#include "../../std.h"

// one more hack
#define WITH_NOGLOBAL
#define SOAP_FMAC3 static
#include "flo_soapC.cpp"
#include "flo_soapClient.cpp"
#include "flo_soapServer.cpp"

#include "../../misc/thread.h"
#include "../../misc/escaped.h"
#include "../../misc/log_time.h"
#include "service.h"

#define __FLO_INTERNAL_INCLUDE__
#include "flo.h"

//#include "file_soap.nsmap"
extern SOAP_NMAC struct Namespace flo_soap_namespaces[];

// ---------------------------------------------

class FlOThread: public Thread {
 private:
  FlOService& service_;
  virtual void func(void);
 public:
  FlOThread(FlOService& srv);
};

FlOThread::FlOThread(FlOService& srv):service_(srv),Thread() {
}

void FlOThread::func(void) {
  odlog(3)<<"FlOThread: func"<<endl;
  for(;;) {
    odlog(3)<<"FlOThread: func: sleep 1"<<endl;
    sleep(30);
    odlog(3)<<"FlOThread: func: maintain"<<endl;
    service_.maintain();
    odlog(3)<<"FlOThread: func: sleep 2"<<endl;
    sleep(30);
  };
}

// ------------------------------------------------------------------
// SOAP service

bool flo_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  list<string> register_;
  list<string> notify_;
  list<string> object_;
  string path_;
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    if(*p == '#') continue;
    std::string command;
    int n = input_escaped_string(p,command);
    if(!n) continue;
    p+=n;
    if(command == "register") { 
      std::string url;
      n=input_escaped_string(p,url);
      if(n != 0) {
        register_.push_back(url);
      };
    } else if(command == "path") {
      n=input_escaped_string(p,path_);
    } else if(command == "storage") {
      n=input_escaped_string(p,path_);
    } else if(command == "notify") {
      std::string url;
      n=input_escaped_string(p,url);
      if(n != 0) {
        notify_.push_back(url);
      };
    } else if(command == "object") {
      std::string id;
      n=input_escaped_string(p,id); p+=n;
      if(n != 0) {
        object_.push_back(id);
      };
    } else if(command == "end") {
      break;
    } else {
      odlog(0)<<"FlO(configure): skipping unknown command: "<<command<<endl;
    };
  };
  FlOService* srv = new FlOService(uri,path_);
  for(list<string>::iterator i = object_.begin();i!=object_.end();++i) {
    pFlOObject o = new FlOObject(*srv,*i,FlOObject::ordinary);
  };
  for(list<string>::iterator i = notify_.begin();i!=notify_.end();++i) {
    pFlOServiceRemote s = new FlOServiceRemote(*i,"");
    srv->add(s);
  };
  for(list<string>::iterator i = register_.begin();i!=register_.end();++i) {
    pFlOServiceRemote s = new FlOServiceRemote(*i,"");
    srv->registr(s);
  };
  FlOThread* thr = new FlOThread(*srv);
  prop.arg=srv;
  prop.subtree=false;
  thr->attention();
  return true;
}

HTTP_Service* flo_service_creator(HTTPS_Connector& c,const char* uri,void* arg){
  FlOService* srv = (FlOService*)arg;
  if(srv == NULL) return NULL;
  return new HTTP_FlO(&c,*srv);
}

#ifdef OLD_GSOAP
#error gSOAP older than 2.5.2 is not supported.
#else
void HTTP_FlO::soap_methods(void) {
  odlog(3)<<"soap_methods: tag: "<<sp.tag<<endl;
  if(!soap_match_tag(&sp,sp.tag,"nf:notification")) soap_serve_nf__notification(&sp);
  else sp.error = SOAP_NO_METHOD;
}
#endif

HTTP_FlO::HTTP_FlO(HTTPS_Connector* c,FlOService& srv):HTTP_ServiceAdv(c),s(srv) {
  soap_init();
  sp.namespaces=flo_soap_namespaces;
  sp.user=this;
}

HTTP_FlO::~HTTP_FlO(void) {
  soap_deinit();
}

HTTP_Error HTTP_FlO::get(const char* uri,int &keep_alive) {
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_FlO::put(const char* uri,int &keep_alive) {
  return HTTP_NOT_IMPLEMENTED;
}

int nf__notification(struct soap* sp,int __size_msg,char** msg,struct nf__response &r) {
  odlog(3)<<"nf__notification"<<endl;
  HTTP_FlO* it = (HTTP_FlO*)(sp->user);
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  odlog(3)<<"nf__notification: size: "<<__size_msg<<endl;
  if(__size_msg <= 0) return SOAP_OK;
  std::string dn = it->c->identity_subject();
  odlog(3)<<"nf__notification: dn: "<<dn<<endl;
  pFlOServiceRemote srv = it->s.service(dn);
  for(int n = 0;n<__size_msg;n++) {
    string message(msg[n]);
    odlog(3)<<"nf__notification: message: "<<message<<endl;
    FlOMessage* m = FlOMessage::deserialize(message);
    if(m == NULL) {
      r.error_code=ERR_WRONG_INPUT_DATA; continue;
    };
    it->s.notification(*m,srv);
  };
  return SOAP_OK;
}

int nf__registration(struct soap* sp,char* url,char* id,char* filter,struct nf__response &r) {
  odlog(3)<<"nf__registration"<<endl;
  HTTP_FlO* it = (HTTP_FlO*)(sp->user);
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  if(url == NULL) { r.error_code=ERR_WRONG_INPUT_DATA; return SOAP_OK; };
  std::string dn = it->c->identity_subject();
  pFlOServiceRemote rsrv = new FlOServiceRemote(url,dn); 
  it->s.add(rsrv);
  return SOAP_OK;
}

int nf__add(struct soap* sp,int __size_obj,char** obj,struct nf__response &r) {
  odlog(3)<<"nf__add"<<endl;
  HTTP_FlO* it = (HTTP_FlO*)(sp->user);
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  if(obj == NULL) return SOAP_OK;
  if(__size_obj <= 0) return SOAP_OK;
  for(int n = 0;n<__size_obj;n++) {
    string object(obj[n]);
    odlog(3)<<"nf__add: object: "<<object<<endl;
    FlOObject* o = FlOObject::deserialize(object);
    if(o == NULL) {
      r.error_code=ERR_WRONG_INPUT_DATA;
      r.error_description="Object can't be parsed";
      continue;
    };
    pFlOObject p = it->s.find(o->id());
    if(p) {
      r.error_code=ERR_OBJECT_EXISTS;
      r.error_description="Object with such ID already exists";
      continue;
    };
    it->s.add(*o);
  };
  return SOAP_OK;
}

int nf__find(struct soap* sp,char* id,char* filter,struct nf__object_response &r) {
  odlog(3)<<"nf__find"<<endl;
  HTTP_FlO* it = (HTTP_FlO*)(sp->user);
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  r.__size_obj=0; r.obj=NULL;
  if(id == NULL) {
    r.error_code=ERR_MISSING_INPUT_DATA;
    r.error_description="Object's ID is missing";
    return SOAP_OK;
  };
  pFlOObject p = it->s.find(string(id));
  if(!p) {
    r.error_code=ERR_OBJECT_MISSING;
    r.error_description="No object found with such ID";
    return SOAP_OK;
  };
  string o;
  p->serialize(o);
  char* obj = (char*)soap_malloc(sp,o.length()+1);
  r.obj=(char**)soap_malloc(sp,sizeof(char*));
  if((obj == NULL) || (r.obj == NULL)) {
    r.error_code=ERR_INTERNAL_ERROR;
    r.error_description="Failed toallocate memory";
    return SOAP_OK;
  };
  r.obj[0]=obj; strcpy(obj,o.c_str());
  r.__size_obj=1;
  return SOAP_OK;
}

